"""
**********************************************************************
**ftrobopy** - Ansteuerung des fischertechnik TXT Controller in Python
**********************************************************************
(c) 2015 by Torsten Stuehn
"""

import sys
import socket
import struct
import time

__author__      = "Torsten Stuehn"
__copyright__   = "Copyright 2015 by Torsten Stuehn"
__credits__     = "fischertechnik GmbH"
__license__     = ""
__version__     = "0.5"
__maintainer__  = "Torsten Stuehn"
__email__       = "stuehn@mailbox.org"
__status__      = "Prototype"
__date__        = "07/05/2015"

class ftTXT(object):
  """
    Basisklasse zum fischertechnik TXT Computer.
    Implementiert das Protokoll zum Datenaustausch ueber Unix Sockets.
    Die Methoden dieser Klasse werden typischerweise vom End-User nicht direkt aufgerufen, sondern
    nur indirekt ueber die Methoden der Klasse ftrobopy.ftrobopy, die eine Erweiterung der Klasse ftrobopy.ftTXT darstellt.
  """

  C_VOLTAGE    = 0
  C_SWITCH     = 1
  C_RESISTOR   = 2
  C_ULTRASONIC = 3
  C_ANALOG     = 0
  C_DIGITAL    = 1
  C_OUTPUT     = 0
  C_MOTOR      = 0

  def __init__(self, host, port):
    """
      Initialisierung der ftTXT Klasse:
        
      * Alle Ausgaenge werden per default auf 1 (=Motor) gesetzt
      * Alle Eingaenge werden per default auf 1, 0 (=Taster, digital) gesetzt
      * Alle Zaehler werden auf 0 gesetzt
        
      :param host: Hostname oder IP-Nummer des TXT Moduls
      :type host: string
        
      - '127.0.0.1' im Downloadbetrieb
      - '192.168.7.2' im USB Offline-Betrieb
      - '192.168.8.2' im WLAN Offline-Betrieb
      - '192.168.9.2' im Bluetooth Offline-Betreib
        
      :param port: Portnummer (normalerweise 65000)
      :type port: integer
    
      :return: Leer
    
      Anwedungsbeispiel:
    
      >>> import ftrobopy
      >>> txt = ftrobopy.ftTXT('192.168.7.2', 65000)
    """
    
    self._host=host
    self._port=port #: Portnummer (normalerweise 65000)
    self._sock=socket.socket()
    try:
      self._sock.connect((host, port))
    except:
      print 'Connection to ', host, ':',port,' failed.'
      sys.exit()
    self._camera_sock   = None
    self._camera_online = False
    self._update_timer  = time.time()
    self._sound_timer   = self._update_timer
    self._sound_length  = 0
    ##########################################
    # queryStatus data
    ##########################################
    self._m_devicename = ''
    self._m_version    = ''
    ##########################################
    # updateStatus data
    ##########################################
    self._config_id            = 0
    self._m_extension_id       = 0
    self._ftX1_pgm_state_req   = 0
    self._ftX1_old_FtTransfer  = 0
    self._ftX1_dummy           = '\x00\x00'
    # Configuration of motors
    # 0=single output O1/O2
    # 1=motor output M1
    self._ftX1_motor           = [1,1,1,1]
    # Universal input mode, see enum InputMode:
    # MODE_U=0
    # MODE_R=1
    # MODE_R2=2
    # MODE_ULTRASONIC=3
    # MODE_INVALID=4
    self._ftX1_uni            = [1,1,'\x00\x00',
                                 1,1,'\x00\x00',
                                 1,1,'\x00\x00',
                                 1,1,'\x00\x00',
                                 1,1,'\x00\x00',
                                 1,1,'\x00\x00',
                                 1,1,'\x00\x00',
                                 1,1,'\x00\x00'] # [ UINT8 mode, BOOL8 digital, char dummy[2] ] 
    # 0=normal, 1=inverted (not really used)
    self._ftX1_cnt            = [1,'\x00\x00\x00',
                                 1,'\x00\x00\x00',
                                 1,'\x00\x00\x00',
                                 1,'\x00\x00\x00'] # [ UINT8 mode, char dummy[3] ]
    # additional motor configuration data (currently not used)
    self._ftX1_motor_config   = [0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0] # INT16[4][4]
    ##########################################
    # exchangeConfig Data
    ##########################################
    self._pwm          = [0,0,0,0,0,0,0,0]
    self._motor_sync   = [0,0,0,0]
    self._motor_dist   = [0,0,0,0]
    self._motor_cmd_id = [0,0,0,0]
    self._last_motor_cmd_id = self._motor_cmd_id
    self._counter      = [0,0,0,0]
    self._sound        = 0
    self._sound_index  = 0
    self._sound_repeat = 0
    ##########################################
    # response from last exchangeConfig Data
    ##########################################
    self._current_input          = [0,0,0,0,0,0,0,0]
    self._current_counter        = [0,0,0,0]
    self._current_counter_value  = [0,0,0,0]
    self._current_counter_cmd_id = [0,0,0,0]
    self._current_motor_cmd_id   = [0,0,0,0]
    self._current_sound_cmd_id   = 0
    self._current_ir             = range(26)

  def queryStatus(self):
    """
       Abfrage des Geraetenamens und der Firmware Versionsnummer des TXT
       Nach dem Umwandeln der Versionsnummer in einen Hexadezimalwert, kann
       die Version direkt abgelesen werden.
       
       :return: Geraetename (string), Versionsnummer (integer)
       
       Anwendungsbeispiel:
       
       >>> name, version = txt.queryStatus()
    """
    m_id         = 0xDC21219A
    m_resp_id    = 0xBAC9723E
    buf          = struct.pack('<I', m_id)
    res          = self._sock.send(buf)
    data         = self._sock.recv(512)
    fstr         = '<I16sI'
    response_id  = 0
    if len(data) == struct.calcsize(fstr):
      response_id, m_devicename, m_version = struct.unpack(fstr, data)
    else:
      m_devicename = ''
      m_version    = ''
    if response_id != m_resp_id:
      print 'WARNING: ResponseID ', hex(response_id),'of queryStatus command does not match'
    self._m_devicename = m_devicename
    self._m_version    = m_version
    return m_devicename, m_version

  def getDevicename(self):
    """
       Liefert den zuvor mit queryStatus() ausgelesenen Namen des TXT zurueck
       
       :return: Geraetename (string)
       
       Anwendungsbeispiel:
       
       >>> print 'Name des TXT: ', txt.getDevicename()
    """
    return self._m_devicename

  def getVersion(self):
    """
       Liefert die zuvor mit queryStatus() ausgelesene Versionsnummer zurueck.
       Um die Firmwareversion direkt ablesen zu koennen, muss diese Nummer noch in
       einen Hexadezimalwert umgewandelt werden
       
       :return: Versionsnummer (integer)
       
       Anwendungsbeispiel:
       
       >>> print hex(txt.getVersion())
    """
    return self._m_version
    
  def startOnline(self):
    """
       Startet den Onlinebetrieb des TXT
       
       :return: Leer
       
       Anwendungsbeispiel:
       
       >>> txt.startOnline()
    """
    m_id       = 0x163FF61D 
    m_resp_id  = 0xCA689F75
    buf        = struct.pack('<I64s', m_id,'')
    res        = self._sock.send(buf)
    data       = self._sock.recv(512)
    fstr       = '<I'
    response_id = 0
    if len(data) == struct.calcsize(fstr):
      response_id, = struct.unpack(fstr, data)
    if response_id != m_resp_id:
      print 'WARNING: ResponseID ', hex(response_id),' of startOnline command does not match'
    return None

  def stopOnline(self):
    """
       Beendet den Onlinebetrieb des TXT
       
       :return: Leer
       
       Anwendungsbeispiel:
       
       >>> txt.stopOnline()
    """
    m_id       = 0x9BE5082C
    m_resp_id  = 0xFBF600D2
    buf        = struct.pack('<I', m_id)
    res        = self._sock.send(buf)
    data       = self._sock.recv(512)
    fstr       = '<I'
    response_id = 0
    if len(data) == struct.calcsize(fstr):
      response_id, = struct.unpack(fstr, data)
    if response_id != m_resp_id:
      print 'WARNING: ResponseID ', hex(response_id), ' of stopOnline command does not match'
    return None

  def setConfig(self, M, I):
    """
       Einstellung der Konfiguration der Ein- und Ausgaenge des TXT.
       Diese Funktion setzt nur die entsprechenden Werte in der ftTXT-Klasse.
       Zur Uebermittlung der Werte an den TXT wird die updateConfig-Methode verwendet.
       
       :param M: Konfiguration der 4 Motorausgaenge (0=einfacher Ausgang, 1=Motorausgang)
       :type M: int[4]

       - Wert=0: Nutzung der beiden Ausgaenge als einfache Outputs
       - Wert=1: Nutzung der beiden Ausgaenge als Motorausgang (links-rechts-Lauf)
 
       :param I: Konfiguration der 8 Eingaenge
       :type I: int[8][2]
       
       :return: Leer
       
       Anwendungsbeispiel:
       
       - Konfiguration der Ausgaenge M1 und M2 als Motorausgaenge
       - Konfiguration der Ausgaenge O5/O6 und O7/O8 als einfache Ausgaenge
       
       
       - Konfiguration der Eingaenge I1, I2, I6, I7, I8 als Taster
       - Konfiguration des Eingangs I3 als Ultraschall Entfernungsmesser
       - Konfiguration des Eingangs I4 als analoger Spannungsmesser
       - Konfiguration des Eingangs I5 als analoger Widerstandsmesser
       
       >>> M = [txt.C_MOTOR, txt.C_MOTOR, txt.C_OUTPUT, txt.C_OUTPUT]
       >>> I = [(txt.C_SWITCH,     txt.C_DIGITAL),
                (txt.C_SWITCH,     txt.C_DIGITAL),
                (txt.C_ULTRASONIC, txt.C_ANALOG),
                (txt.C_VOLTAGE,    txt.C_ANALOG),
                (txt.C_RESISTOR,   txt.C_ANALOG),
                (txt.C_SWITCH,     txt.C_DIGITAL),
                (txt.C_SWITCH,     txt.C_DIGITAL),
                (txt.C_SWITCH,     txt.C_DIGITAL)]
       >>> txt.setConfig(M, I)
       >>> txt.updateConfig()
    """
    m_id       = 0x060EF27E
    m_resp_id  = 0x9689A68C
    self._config_id += 1
    # Configuration of motors
    # 0=single output O1/O2
    # 1=motor output M1
    # self.ftX1_motor          = [M[0],M[1],M[2],M[3]]  # BOOL8[4]
    self._ftX1_motor            = M
    # Universal input mode, see enum InputMode:
    # MODE_U=0
    # MODE_R=1
    # MODE_R2=2
    # MODE_ULTRASONIC=3
    # MODE_INVALID=4
    # print "setConfig I=", I
    self._ftX1_uni           = [I[0][0],I[0][1],'\x00\x00',
                                I[1][0],I[1][1],'\x00\x00',
                                I[2][0],I[2][1],'\x00\x00',
                                I[3][0],I[3][1],'\x00\x00',
                                I[4][0],I[4][1],'\x00\x00',
                                I[5][0],I[5][1],'\x00\x00',
                                I[6][0],I[6][1],'\x00\x00',
                                I[7][0],I[7][1],'\x00\x00'] # [ UINT8 mode, BOOL8 digital, char dummy[2] ] 
    return None

  def getConfig(self):
    """
       Abfrage der aktuellen Konfiguration des TXT
       
       :return: M[4], I[8][2]
       :rtype: M:int[4], I:int[8][2]
       
       Anwendungsbeispiel: Aenderung des Eingangs I2 auf analoge Ultraschall Distanzmessung

       - Hinweis: Feldelemente werden in Python typischerweise durch die Indizes 0 bis N-1 angesprochen
       - Der Eingang I2 des TXT wird in diesem Beispiel ueber das Feldelement I[1] angesprochen
       
       >>> M, I = txt.getConfig()
       >>> I[1] = (txt.C_ULTRASONIC, txt.C_ANALOG)
       >>> txt.setConfig(M, I)
       >>> txt.updateConfig()
    """
    m  = self._ftX1_motor
    i  = self._ftX1_uni
    ii = [(i[0], i[1]), (i[3], i[4]),(i[6], i[7]),(i[9], i[10]),(i[12], i[13]),(i[15], i[16]),(i[18], i[19]),(i[21], i[22]) ]
    return m, ii 

  def updateConfig(self):
    """
       Uebertragung der Konfigurationsdaten fuer die Ein- und Ausgaenge zum TXT
       
       :return: Leer
       
       Anwendungsbeispiel:
       
       >>> txt.setConfig(M, I)
       >>> txt.updateConfig()
    """
    m_id       = 0x060EF27E
    m_resp_id  = 0x9689A68C
    self._config_id += 1
    fields = [m_id, self._config_id, self._m_extension_id]
    fields.append(self._ftX1_pgm_state_req)
    fields.append(self._ftX1_old_FtTransfer)
    fields.append(self._ftX1_dummy)
    fields += self._ftX1_motor
    fields += self._ftX1_uni
    fields += self._ftX1_cnt
    fields += self._ftX1_motor_config
    buf = struct.pack('<Ihh B B 2s BBBB BB2s BB2s BB2s BB2s BB2s BB2s BB2s BB2s B3s B3s B3s B3s 16h', *fields)
    res = self._sock.send(buf)
    data = self._sock.recv(512)
    fstr    = '<I'
    response_id = 0
    if len(data) == struct.calcsize(fstr):
      response_id, = struct.unpack(fstr, data)
    if response_id != m_resp_id:
      print 'WARNING: ResponseID ', hex(response_id),' of updateConfig command does not match'
    return None

  def exchangeData(self):
    """
       Low-Level Uebermittlung von Motor-, bzw. Output-Daten (z.B. Motorgeschwindigkeit, Motordistanz) an den TXT
       und Abholen der aktuellen Eingangswerte vom TXT.

       :return: Leer
       
       Hinweis
       
       - Diese Methode wird typischerweise nicht direkt aufgerufen, sondern nur indirekt ueber die update()-Methode (siehe unten)
    """
    m_id         = 0xCC3597BA
    m_resp_id    = 0x4EEFAC41
    fields  = [m_id]
    fields += self._pwm
    fields += self._motor_sync
    fields += self._motor_dist
    fields += self._motor_cmd_id
    fields += self._counter
    fields += [self._sound, self._sound_index, self._sound_repeat,0,0]
    #print fields
    buf = struct.pack('<I8h4h4h4h4hHHHbb', *fields)
    res = self._sock.send(buf)
    data = self._sock.recv(512)
    fstr    = '<I8h4h4h4h4hH4bB4bB4bB4bB4bBb'
    response_id = 0
    #print 'received ' ,len(data), 'bytes ', ' (expected: ',struct.calcsize(fstr),' bytes)'
    if len(data) == struct.calcsize(fstr):
      response = struct.unpack(fstr, data)
    else:
      print 'Received data size does not match length of format string'
      return None
    response_id = response[0]
    if response_id != m_resp_id:
      print 'WARNING: ResponseID ', hex(response_id),' of exchangeData command does not match'
    #print response
    self._current_input          = response[1:9]
    self._current_counter        = response[9:13]
    self._current_counter_value  = response[13:17]
    self._current_counter_cmd_id = response[17:21]
    self._current_motor_cmd_id   = response[21:25]
    self._current_sound_cmd_id   = response[25]
    self._current_ir             = response[26:52]
    return None

  def startCameraOnline(self):
    """
      Die Camerafunktionen werden noch nicht unterstuetzt.
    """
    m_id                 = 0x882A40A6
    m_resp_id            = 0xCF41B24E
    self._m_width         = 320
    self._m_height        = 240
    self._m_framerate     = 30
    self._m_powerlinefreq = 1 # 0=auto, 1=50Hz, 2=60Hz
    buf        = struct.pack('<I4i', m_id,
                                     self._m_width,
                                     self._m_height,
                                     self._m_framerate,
                                     self._m_powerlinefreq)
    res        = self._sock.send(buf)
    data       = self._sock.recv(512)
    fstr       = '<I'
    response_id = 0
    if len(data) == struct.calcsize(fstr):
      response_id, = struct.unpack(fstr, data)
    if response_id != m_resp_id:
      print 'WARNING: ResponseID ', hex(response_id),' of startCameraOnline command does not match'
    self._camera_sock=socket.socket(socket.AF_INET, socket.SOCK_STREAM, socket.IPPROTO_TCP)
    try:
      self._camera_sock.connect((self.host, self.port+1))
    except:
      print 'Connection to camera failed'
      # sys.exit()
    else:
      self._camera_online = True
    return None

  def stopCameraOnline(self):
    """
      Die Camerafunktionen werden noch nicht unterstuetzt.
    """
    m_id                 = 0x17C31F2F
    m_resp_id            = 0x4B3C1EB6
    buf        = struct.pack('<I', m_id)
    res        = self._sock.send(buf)
    data       = self._sock.recv(512)
    fstr       = '<I4i'
    response_id = 0
    if len(data) == struct.calcsize(fstr):
      response_id, = struct.unpack(fstr, data)
    if response_id != m_resp_id:
      print 'WARNING: ResponseID ', hex(response_id),' of stopCameraOnline command does not match'
    return None

  def cameraOnlineFrame(self):
    """
      Die Camerafunktionen werden noch nicht unterstuetzt.
    """
    m_id     = 0xBDC2D7A1
    m_ack_id = 0xADA09FBA
    data       = self._camera_sock.recv(100000)
    fstr       = '<Iihhii'
    ds_size    = struct.calcsize(fstr) # data struct size without jpeg data
    jpeg_size  = len(data) - ds_size   # size of jpeg data
    vfstr      = fstr + str(jpeg_size) + 'B'
    response   = struct.unpack(vfstr, data)
    self._m_numframesready      = response[1]
    self._m_framewidth          = response[2]
    self._m_frameheight         = response[3]
    self._m_framesizeraw        = response[4]
    self._m_framesizecompressed = response[5]
    self._m_framedata           = response[6:6+self.frame]
    # print 'current jpeg framesize: ',self._m_framesizecompressed,' waiting frames: ',self._num_framesready
    # send acknowledgement
    buf = struct.pack('<I', m_ack_id)
    ack = self._camera_sock.send(buf)

  def cameraFramesReady(self):
    """
      Die Camerafunktionen werden noch nicht untertuetzt.
    """
    if self.cameraOnline():
      return self._numframesready
    else:
      return 0

  def update(self, wait=False):
    """
      Senden von Motor- und/oder Output Parametern an den TXT und unmittelbares Abholen der
      Werte der TXT Eingaenge.
     
      * der TXT kommuniziert nur alle 20ms (oder 10ms ?) ueber den Port 65000 mit der Anwendung
      * die update()-Methode stellt automatisch sicher, dass dieses 20ms Kommunikationsintervall eingehalten wird
        und die exchangeData()-Methode nicht haeufiger aufgerufen wird. Bei direktem (haeufigen) Aufruf der
        exchangeData()-Methode bricht sonst evtl. die Verbindung zum TXT ab.
      * Diese Funktionsweise wurde "experimentell" ermittelt. Die Dokumentation von fischertechnik ist an dieser Stelle nicht ganz vollstaendig.
     
      :param wait: Warten auf naechstes 20ms Intervall
      :type wait: boolean True oder False
      
      - Wird dieser Parameter auf True gesetzt, wartet die update()-Methode bis zum naechsten 20ms-Zyklus und fuehrt dann den Datenaustausch durch.
      - Steht dieser Parameter auf False (default) wird der Datenaustausch nur durchgefuehrt, falls seit dem letzten Datenabgleich mindestens 20ms
        vergangen sind, falls nicht, wird kein Datenaustausch vorgenommen.
        
      :return: Leer
     
      Anwendungsbeispiel:
      
      - Motorausgang M2 auf volle Geschwindigkeit vorwaerts schalten
      - Eingang I6 abfragen, in der Annahme, dass an I6 ein Taster angeschlossen ist
      
      >>> txt.setPwm(2, 512)
      >>> txt.setPwm(3, 0)
      >>> txt.update(wait=True)
      >>> Taster_I6_gedrueckt = txt.getCurrentInput(6)
    """
    if wait:
      while time.time() - self._update_timer < 0.02:
        if self.cameraFramesReady() > 0:
          self.cameraOnlineFrame()
        else:
          pass
      self.exchangeData()
      self._update_timer=time.time()
    else:
      if time.time() - self._update_timer >= 0.02:
        if self.cameraFramesReady() > 0:
          self.cameraOnlineFrame()
        self.exchangeData()
        self._update_timer=time.time()

  def sleep(self, seconds):
    """
      Wartet die angegebene Anzahl von Sekunden ab, bevor die Programmausfuehrung fortgesetzt wird.
      Waehrend der Wartezeit wird automatisch kontinuierlich die update()-Methode aufgerufen um die Verbindung
      zum TXT aufrecht zu erhalten.
      
      :param seconds: Wartezeit in Sekunden
      :type seconds: float
      
      Anwendungsbeispiel:
      
      Der Programmablauf wird fuer eine halbe Sekunde pausiert (z.B. um einem Pneumatik-Zylinder, der von einem Magnetventil angesteuert wird,
      Zeit zu geben, in seine Endstellung zu kommen)
      
      >>> txt.sleep(0.5)
    """
    sleep_timer=time.time()
    while time.time() - sleep_timer < seconds:
      self.update(wait=True)

  def incrMotorCmdId(self, idx):
    """
      Erhoehung der sog. Motor Command ID um 1.
      
      Diese Methode muss immer dann aufgerufen werden, wenn die Distanzeinstellung eines Motors (gemessen ueber die 4 schnellen Counter-Eingaenge)
      geaendert wurde oder wenn ein Motor mit einem anderen Motor synchronisiert werden soll. Falls nur die Motorgeschwindigkeit veraendert wurde,
      ist der Aufruf der incrMotorCmdId()-Methode nicht notwendig.
      
      :param idx: Nummer des Motorausgangs
      :type idx: integer
      
      Achtung:
      
      * Die Zaehlung erfolgt hier von 0 bis 3, idx=0 entspricht dem Motorausgang M1 und idx=3 entspricht dem Motorausgang M4
      
      Anwendungsbeispiel:
      
      Der Motor, der am TXT-Anschluss M2 angeschlossen ist, soll eine Distanz von 200 (Counterzaehlungen) zuruecklegen.
      
      >>> txt.setMotorDistance(1, 200)
      >>> txt.incrMotorCmdId(1)
      >>> txt.update(wait=True)
    """
    self._motor_cmd_id[idx] += 1
    return None

  def getMotorCmdId(self, idx=None):
    """
      Liefert die letzte Motor Command ID eines Motorausgangs (oder aller Motorausgaenge als array) zurueck.
      
      :param idx: Nummer des Motorausgangs. Falls dieser Parameter nicht angeben wird, wird die Motor Command ID aller Motorausgaenge als Array[4] zurueckgeliefert.
      :type idx: integer
      
      :return: Die Motor Command ID eines oder aller Motorausgaenge
      :rtype: integer oder integer[4] array
      
      Anwendungsbeispiel:
      
      >>> letzte_cmd_id = txt.getMotorCmdId(4)
    """
    if idx != None:
      return self._motor_cmd_id[idx]
    else:
      return self._motor_cmd_id

  def cameraOnline(self):
    """
      Die Camerafunktionen werden noch nicht unterstuetzt
    """
    return self._camera_online

  def getSoundCmdId(self):
    """
      Liefert die letzte Sound Command ID zurueck.
      
      :return: Letzte Sound Command ID
      :rtype: integer
      
      Anwendungsbeispiel:
      
      >>> last_sound_cmd_id = txt.getSoundCmdId()
    """
    return self._sound

  def incrCounterCmdId(self, idx):
    """
      Erhoehung der Counter Command ID um eins.
      Falls die Counter Command ID eines Counters um eins erhoeht wird, wird der entsprechende Counter zurueck auf 0 gesetzt.
      
      :param idx: Nummer des schnellen Countereingangs, dessen Command ID erhoeht werden soll. (Hinweis: die Zaehlung erfolgt hier von 0 bis 3 fuer die Counter C1 bis C4)
      :type idx: integer
      
      :return: Leer
      
      Anwendungsbeispiel:
      
      Erhoehung der Counter Command ID des Counters C4 um eins.
      
      >>> txt.incrCounterCmdId(3)
    """
    self._counter[idx] += 1
    return None

  def incrSoundCmdId(self):
    """
      Erhoehung der Sound Command ID um eins.
      Die Sound Command ID muss immer dann um eins erhoeht werden, falls ein neuer Sound gespielt werden soll oder
      wenn die Wiederholungsanzahl eines Sounds veraendert wurde. Falls kein neuer Sound index gewaehlt wurde und auch
      die Wiederholungsrate nicht veraendert wurde, wird der aktuelle Sound erneut abgespielt.
      
      :return: Leer
      
      Anwendungsbeispiel:
      
      >>> txt.incrSoundCmdId()
    """
    self._sound += 1
    return None

  def setSoundIndex(self, idx):
    """
      Einstellen eines neuen Sounds.
      
      :param idx: Nummer des neuen Sounds (0=Kein Sound, 1-29 Sounds des TXT)
      :type idx: integer
      
      :return: Leer
      
      Anwendungsbeispiel:
      
      Sound "Augenzwinkern" einstellen und 2 mal abspielen.
      
      >>> txt.setSoundIndex(26)
      >>> txt.setSoundRepeat(2)
      >>> txt.incrSoundCmdId()
    """
    self._sound_index = idx
    return None

  def getSoundIndex(self):
    """
      Liefert die Nummer des aktuell eingestellten Sounds zurueck.
      
      :return: Nummer des aktuell eingestellten Sounds
      :rtype: integer
      
      Anwendungsbeispiel:
      
      >>> aktueller_sound = txt.getSoundIndex()
    """
    return self._sound_index

  def setSoundRepeat(self, rep):
    """
      Einstellen der Anzahl der Wiederholungen eines Sounds.
      
      :param rep: Anzahl der Wiederholungen (0=unendlich oft wiederholen)
      :type rep: integer
      
      Anwendungsbeispiel:
      
      "Motor-Sound" unendlich oft (d.h. bis zum Ende des Programmes oder bis zur naechsten Aenderung der Anzahl der Wiederholungen) abspielen.
      
      >>> txt.setSound(19) # 19=Motor-Sound
      >>> txt.setSoundRepeat(0)
    """
    self._sound_repeat = rep
    return None

  def getSoundRepeat(self):
    """
      Liefert die aktuell eingestellte Wiederholungs-Anzahl des Sounds zurueck.
      
      :return: Aktuell eingestellte Wiederholungs-Anzahl des Sounds.
      :rtype: integer
      
      Anwendungsbeispiel:
      
      >>> repeat_rate = txt.getSoundRepeat()
    """
    return self._sound_repeat

  def getCounterCmdId(self, idx=None):
    """
      Liefert die letzte Counter Command ID eines (schnellen) Counters zurueck
      
      :param idx: Nummer des Counters. (Hinweis: die Zaehlung erfolgt hier von 0 bis 3 fuer die Counter C1 bis C4)
      
      Anwendungsbeispiel:
      
      Counter Command ID des schnellen Counters C3 in Variable num einlesen.
      
      >>> num = txt.getCounterCmdId(2)
    """
    if idx != None:
      return self._counter[idx]
    else:
      return self._counter

  def setPwm(self, idx, value):
    """
      Einstellen des Ausgangswertes fuer einen Motor- oder Output-Ausgang.
      
      :param idx: Nummer des Ausgangs. (Hinweis: die Zaehlung erfolgt hier von 0 bis 7 fuer die Ausgaenge O1-O8)
      :type idx: integer (0-7)
      
      :param value: Wert, auf den der Ausgang gesetzt werden soll (0:Ausgang ausgeschaltet, 512: Ausgang auf maximum)
      :type value: integer (0-512)
      
      :return: Leer
      
      Anwendungsbeispiel:
      
      * Motor am Anschluss M1 soll mit voller Geschwindigkeit Rueckwaerts laufen.
      * Lampe am Anschluss O3 soll mit halber Leuchtkraft leuchten.
      
      >>> txt.setPwm(0,0)
      >>> txt.setPwm(1,512)
      >>> txt.setPwm(2,256)
      >>> txt.update(wait=True) # uebermitteln der Daten an den TXT
    """
    self._pwm[idx]=value
    return None

  def getPwm(self, idx=None):
    """
      Liefert die zuletzt eingestellten Werte der Ausgaenge O1-O8 (als array[8]) oder den Wert eines Ausgangs.
      
      :param idx:
      :type idx: integer oder None, bzw. leer
      
      - Wenn kein idx-Parameter angegeben wurde, werden alle Pwm-Einstellungen als array[8] zurueckgeliefert.
      - Ansonsten wird nur der Pwm-Wert des mit idx spezifizierten Ausgangs zurueckgeliefert.

      Hinweis: der idx-Parameter wird angeben von 0 bis 7 fuer die Ausgaenge O1-O8
                  
      :return: der durch (idx+1) spezifizierte Ausgang O1 bis O8 oder das gesamte Pwm-Array
      :rtype: integer oder integer array[8]
                  
      Anwendungsbeispiel:
      
      Liefert die
      
      >>> M1_a = txt.getPwm(0)
      >>> M1_b = txt.getPwm(1)
      >>> if M1_a > 0 and M1_b == 0:
            print "Der an M1 angeschlossene Motor laeuft mit der Geschwindigkeit ", M1_a, " vorwaerts."
          else:
            if M1_a == and M1_b > 0:
              print "Der an M1 angeschlossene Motor laeuft mit der Geschwindigkeit ", M1_b, " rueckwaerts."
    """
    if idx != None:
      return self._pwm[idx]
    else:
      return self._pwm

  def setMotorSyncMaster(self, idx, value):
    """
      Hiermit koennen zwei Motoren miteinander synchronisiert werden, z.B. fuer perfekten Geradeauslauf.
      
      :param idx: Der Motorausgang, der synchronisiert werden soll
      :type idx: integer
      
      :param value: Die Numer des Motorausgangs, mit dem synchronisiert werden soll.
      :type value: integer
      
      :return: Leer
      
      Hinweis:
      
      - der idx-Parameter wird angeben von 0 bis 3 fuer die Motor-Ausgaenge M1 bis M4.
      - der value-PArameter wird angeben von 1 bis 4 fuer die Motor-Ausgaenge M1 bis M4.
      
      Anwendungsbeispiel:
      
      Die Motorausgaenge M1 und M2 werden synchronisiert.
      Um die Synchronisations-Befehle abzuschliessen, muss ausserdem die MotorCmdId der Motoren erhoeht werden.
      
      >>> txt.setMotorSyncMaster(0, 2)
      >>> txt.setMotorSyncMaster(1, 1)
      >>> txt.incrMotorCmdId(0)
      >>> txt.incrMotorCmdId(1)
      >>> txt.update(Wait=True)
    """
    self._motor_sync[idx]=value
    return None

  def getMotorSyncMaster(self, idx=None):
    """
      Liefert die zuletzt eingestellte Konfiguration der Motorsynchronisation fuer einen oder alle Motoren zurueck.
      
      :param idx: Die Nummer des Motors, dessen Synchronisation geliefert werden soll oder None oder <leer> fuer alle Ausgaenge.
      :type idx: integer

      :return: Leer

      Hinweis:
      
      - der idx-Parameter wird angeben von 0 bis 3 fuer die Motor-Ausgaenge M1 bis M4.
      - oder None oder <leer> fuer alle Motor-Ausgaenge.
     
      Anwendungsbeispiel:
      
      >>> print "Aktuelle Konfiguration aller Motorsynchronisationen: ", txt.getMotorSyncMaster()
    """
    if idx != None:
      return self._motor_sync[idx]
    else:
      return self._motor_sync

  def setMotorDistance(self, idx, value):
    """
      Hiermit kann die Distanz (als Anzahl von schnellen Counter-Zaehlungen) fuer einen Motor eingestellt werden.
      
      :param idx: Nummer des Motorausgangs
      :type idx: integer

      :return: Leer

      Hinweis:
      
      - der idx-Parameter wird angeben von 0 bis 3 fuer die Motor-Ausgaenge M1 bis M4.

      Anwendungsbeispiel:
      
      Der Motor an Ausgang M3 soll 100 Counter-Zaehlungen lang drehen.
      Um den Distanz-Befehl abzuschliessen, muss ausserdem die MotorCmdId des Motors erhoeht werden.
      
      >>> txt.setMotorDistance(2, 100)
      >>> txt.incrMotorCmdId(2)
      >>> txt.update(wait=True)
    """
    self._motor_dist[idx]=value
    return None

  def getMotorDistance(self, idx=None):
    """
      Liefert die zuletzt eingestellte Motordistanz fuer einen oder alle Motorausgaenge zurueck.
      
      :param idx: Nummer des Motorausgangs
      :type idx: integer

      :return: Letzte eingestellte Distanz eines Motors (idx=0-3) oder alle zuletzt eingestellten Distanzen (idx=None oder kein idx-Parameter angegeben)

      Hinweis:

      - der idx-Parameter wird angeben von 0 bis 3 fuer die Motor-Ausgaenge M1 bis M4.

      Anwendungsbeispiel:
      
      >>> print "Die zuletzt mit setMotorDistance() eingestellte Distanz fuer den Motorausgang M2 lautet: ", txt.getMotorDistance(1)
    """
    if idx != None:
      return self._motor_dist[idx]
    else:
      return self._motor_dist

  def getCurrentInput(self, idx=None):
    """
       Liefert den aktuellen vom TXT zurueckgelieferten Wert eines Eingangs oder aller Eingaenge als Array

       :param idx: Nummer des Eingangs
       :type idx: integer
       
       :return: Aktueller Wert eines Eingangs (idx=0-7) oder alle aktuellen Eingangswerte des TXT-Controllers als Array[8] (idx=None oder kein idx angegeben)
       
       Hinweis:
       
       - der idx-Parameter wird angeben von 0 bis 7 fuer die Eingaenge I1 bis I8.
       
       Anwendungsbeispiel:
       
       Vor der Abfrage werden die aktuellen Werte beim TXT "abgeholt".
       
       >>> txt.update(wait=True)
       >>> print "Der aktuelle Wert des Eingangs I4 ist: ", txt.getCurrentInput(3)
    """
    if idx != None:
      return self._current_input[idx]
    else:
      return self._current_input

  def getCurrentCounterInput(self, idx=None):
    """
      Zeigt an, ob ein Counter oder alle Counter (als Array[4]) sich seit der letzten Abfrage veraendert haben.
      
      :param idx: Nummer des Counters
      :type idx: integer
        
      :return: Aktueller Status-Wert eines Counters (idx=0-3) oder aller schnellen Counter des TXT-Controllers als Array[4] (idx=None oder kein idx angegeben)
        
      Hinweis:
        
      - der idx-Parameter wird angeben von 0 bis 3 fuer die Counter C1 bis C4.
        
      Anwendungsbeispiel:
        
      Vor der Abfrage werden die aktuellen Werte beim TXT "abgeholt".
        
      >>> txt.update(wait=True)
      >>> c = txt.getCurrentCounterInput(0)
      >>> if c==0:
      >>>   print "Der Counter von Motor M1 hat sich seit der letzten Abfrage nicht veraendert"
      >>> else:
      >>>   print "Der Counter von Motor M1 hat sich seit der letzten Abfrage veraendert"
    """
    if idx != None:
      return self._current_counter[idx]
    else:
      return self._current_counter

  def getCurrentCounterValue(self, idx=None):
    """
      Liefert den aktuellen Wert eines oder aller schnellen Counter Eingaenge zurueck.
      Damit kann z.B. nachgeschaut werden, wie weit ein Motor schon gefahren ist.
      
      :param idx: Nummer des Counters
      :type idx: integer
          
      :return: Aktueller Wert eines Counters (idx=0-3) oder aller schnellen Counter des TXT-Controllers als Array[4] (idx=None oder kein idx angegeben)
          
      Hinweis:
          
      - der idx-Parameter wird angeben von 0 bis 3 fuer die Counter C1 bis C4.
          
      Anwendungsbeispiel:
          
      Vor der Abfrage werden die aktuellen Werte beim TXT "abgeholt".
          
      >>> txt.update(wait=True)
      >>> print "Der aktuelle Wert des Counters von Motor M1 ist: ", txt.getCurrentCounterValue(0)
    """
    if idx != None:
      return self._current_counter_value[idx]
    else:
      return self._current_counter_value

  def getCurrentCounterCmdId(self, idx=None):
    """
      Liefert die aktuelle Counter Command ID eines oder aller Counter zurueck.
      
      :param idx: Nummer des Counters
      :type idx: integer
        
      :return: Aktuelle Commmand ID eines Counters (idx=0-3) oder aller Counter des TXT-Controllers als Array[4] (idx=None oder kein idx angegeben)
        
      Hinweis:
        
      - der idx-Parameter wird angeben von 0 bis 3 fuer die Counter C1 bis C4.
        
      Anwendungsbeispiel:
      
      Vor der Abfrage werden die aktuellen Werte beim TXT "abgeholt".
      
      >>> txt.update(wait=True)
      >>> print "Die aktuelle Counter Command ID von Counter C4 lautet: ", txt.getCurrentCounterCmdId(3)
    """
    if idx != None:
      return self._current_counter_cmd_id[idx]
    else:
      return self._current_counter_cmd_id

  def getCurrentMotorCmdId(self, idx=None):
    """
      Liefert die aktuelle Motor Command ID eines oder aller Motoren zurueck.
          
      :param idx: Nummer des Motors
      :type idx: integer
          
      :return: Aktuelle Commmand ID eines Motors (idx=0-3) oder aller Motoren des TXT-Controllers als Array[4] (idx=None oder kein idx angegeben)
          
      Hinweis:
          
      - der idx-Parameter wird angeben von 0 bis 3 fuer die Motoren M1 bis M4.
          
      Anwendungsbeispiel:
          
      Vor der Abfrage werden die aktuellen Werte beim TXT "abgeholt".
          
      >>> txt.update(wait=True)
      >>> print "Die aktuelle Motor Command ID von Motor M4 lautet: ", txt.getCurrentMotorCmdId(3)
    """
    if idx != None:
      return self._current_motor_cmd_id[idx]
    else:
      return self._current_motor_cmd_id
  
  def getCurrentSoundCmdId(self):
    """
      Liefert die aktuelle Sound Command ID zurueck.
      
      :return: Die aktuelle Sound Command ID
      :rtype: integer
      
      Anwendungsbeispiel:
      
      Vor der Abfrage werden die aktuellen Werte beim TXT "abgeholt".
      
      >>> txt.update(wait=True)
      >>> print "Die aktuelle Sound Command ID ist: ", txt.getCurrentSoundCmdId()
    """
    return self._current_sound_cmd_id

  def getCurrentIr(self):
    """
      Liefert die aktuellen Werte der Infrarot Fernsteuerung zurueck (als Array oder als einzelnen Wert)
      
      Die Werte werden fuer jede einzelne Einstellung der DIP-Switche auf der IR-Fernsteuerung zurueckgeliefert,
      so dass insgesamt 4 Fernsteuerungen abgefragt werden koennen (als Python-Liste von Python-Listen).
      Die 5. Liste innerhalb der Python-Liste gibt den Zustand einer Fernsteuerung mit beliebiger DIP-Switch-Einstellung zurueck.
      Die Einstellung der Switche selbst wird auch zurueckgelifert, aber nicht, wenn wenn die Switche sich aendern, sondern nur, wenn einer der anderen Werte sich aendert.
      
      :param idx: Nummer des IR Eingangs
      :type idx: integer
      
      :return: Aktueller Wert eines IR Eingangs (idx=0-5) oder aller IR Eingaenge des TXT-Controllers als Array[5] (idx=None oder kein idx angegeben)
      
      Hinweis:
      
      - der idx-Parameter wird angeben von 0 bis 5 fuer die einzelnen Listen innerhalb der Liste.
      
      Anwendungsbeispiel:
      
      Vor der Abfrage werden die aktuellen Werte beim TXT "abgeholt".
     
      >>> txt.update(wait=True)
      >>> print "Die aktuellen Werte der IR Fernsteuerung lauten: ", txt.getCurrentIr()
    """
    return self._current_ir


class ftrobopy(ftTXT):
  """
    Erweiterung der Klasse ftrobopy.ftTXT. In dieser Klasse werden verschiedene fischertechnik Elemente
    auf einer hoeheren Abstraktionsstufe (aehnlich den Programmelementen aus der ROBOPro Software) fuer den End-User zur Verfuegung gestellt.
    Derzeit sind die folgenden Programmelemente implementiert:
    
    * **motor**, zur Ansteuerung der Motorausgaenge M1-M4
    * **output**, zur Ansteuerung der universellen Ausgaenge O1-O8
    * **input**, zum Einlesen von Werten der Eingaenge I1-I8
    * **ultrasonic**, zur Bestimmung von Distanzen mit Hilfe des Ultraschall Moduls
    
    Ausserdem werden die folgenden Sound-Routinen zur Verfuegung gestellt:
    
    * **play_sound**
    * **stop_sound**
    * **sound_finished**
    
  """
  def __init__(self, host, port):
    """
      Initialisierung der ftrobopy Klasse:
      
      * Herstellen der Socket-Verbindung zum TXT Controller und Abfrage des Geraetenamens und der Firmwareversionsnummer
      * Starten des Online-Modus des TXT
      * Initialisierung aller Datenfelder der ftTXT Klasse mit Defaultwerten

      :param host: Hostname oder IP-Nummer des TXT Moduls
      :type host: string
      
      - '127.0.0.1' im Downloadbetrieb
      - '192.168.7.2' im USB Offline-Betrieb
      - '192.168.8.2' im WLAN Offline-Betrieb
      - '192.168.9.2' im Bluetooth Offline-Betreib
      
      :param port: Portnummer (normalerweise 65000)
      :type port: integer
      
      :return: Leer
      
      Anwedungsbeispiel:
      
      >>> import ftrobopy
      >>> ftrob = ftrobopy.ftrobopy('192.168.7.2', 65000)
    """
    ftTXT.__init__(self, host, port)
    self.queryStatus()
    print 'Connected to ', self.getDevicename(), ' version ', hex(self.getVersion())
    self.startOnline()
    self.updateConfig()
    for i in range(8):
      self.setPwm(i,0)
    self.exchangeData()

  def __del__(self):
      # if self.cameraOnline():
      #   self.stopCameraOnline()
      # self.stopOnline()
      pass

  def motor(self, output):
    """
      Diese Funktion erzeugt ein Motor-Objekt, das zur Ansteuerung eines Motors verwendet werden kann,
      der an einem der Motorausgaenge M1-M4 des TXT angeschlossen ist.
      
      Anwendungsbeispiel:
      
      >>> Motor1 = ftrob.motor(1)
      
      Das so erzeugte Motor-Objekt hat folgende Methoden:
      
      **setSpeed** (speed)
      
      Einstellung der Motorgeschwindigkeit
      
      :param speed: 
      :type speed: integer
      
      :return: Leer
      
      Gibt an, mit welcher Geschwindigkeit der Motor laufen soll:
      
      - der Wertebereich der Geschwindigkeit liegt zwischen 0 (Motor anhalten) und 512 (maximale Geschwindigkeit)
      - Falls die Geschwindigkeit negativ ist, laeuft der Motor Rueckwaerts
      
      Hinweis: Der eingegebene Wert fuer die Geschwindigkeit haengt nicht linear mit der tatsaechlichen
               Drehgeschwindig des Motors zusammen, d.h. die Geschwindigkeit bei 400 ist nicht doppelt
               so gross, wie die Geschwindigkeit 200.  Bei Hoeheren Werten von speed kann dadurch die
               Geschwindigkeit in feineren Stufen reguliert werden.
      
      Anwendungsbeispiel:
      
      Dieser Befehl wird erst dann ausgefuehrt, wenn das "update(wait=True)"
      Kommando ausgefuehrt wurde. Dadurch ist es moeglich die Geschwindigkeit von mehreren Motoren
      anzugeben und erst dann diese Werte gemeinsam an den TXT zu uebermitteln.
  
      >>> Motor1.setSpeed(512)
      >>> ftrob.update(wait=True)
      
      **setDistance** (distance, syncto=None)
      
      Einstellung der Mototdistanz, die ueber die schnellen Counter gemessen wird.
      
      :param distance: Gibt an, um wieviele Counter-Zaehlungen sich der Motor drehen soll.
      :type distance: integer
      
      :param syncto: Hiermit koennen zwei Motoren synchronisiert werden um z.B. perfekten Geradeauslauf
                     zu ermoeglichen. Als Parameter wird hier ein anderes Motorobjekt uebergeben.
      :type syncto: ftrobopy.motor Objekt
      
      :return: Leer
      
      Anwendungsbeispiel:
      
      Der Motor wir mit dem Motor synchronisiert, der am Motorausgang M2 angeschlossen ist.
      Grundsaetzlich wird dieser Befehl erst dann ausgefuehrt, wenn das "update(wait=True)"
      Kommando ausgefuehrt wurde. Nur so ist moeglich, zwei Motoren exakt zu synchronisieren.
      (das incrMotorCmdId()-Kommando wird innerhalb der setDistance-Routine automatisch ausgefuehrt).

      >>> Motor_links=ftrob.motor(1)
      >>> Motor_rechts=ftrob.motor(2)
      >>> Motor_links.setDistance(100, syncto=Motor_rechts)
      >>> Motor_rechts.setDistance(100, syncto=Motor_rechts)
      >>> ftrob.update(wait=True)
      
      **finished** ()
      
      Abfrage, ob die eingestellte Distanz bereits erreicht wurde.
      
      :return: False: Motor laeuft noch, True: Distanz erreicht
      :rtype: boolean
      
      Anwendungsbeispiel:
      
      Hinweis: Der finished() Befehl holt sich autmatisch zuerst die aktuellen Werte vom TXT
               per ftrob.update(wait=True)
      
      >>> while not Motor1.finished():
            print "Motor laeuft noch"
            
      **stop** ()
      
      Anhalten des Motors durch setzen der Geschwindigkeit auf 0.
      
      :return: Leer
      
      Anwendungsbeispiel:
      
      >>> Motor1.stop()
      >>> ftrob.update(wait=True)
    """
    class mot(object):
      def __init__(self, outer, output):
        self._outer=outer
        self._output=output
        self._speed=0
        self._distance=0
        self._outer.updateConfig()
        self.setSpeed(0)
        self.setDistance(0)
      def setSpeed(self, speed):
        self._speed=speed
        if speed > 0:
         self._outer.setPwm((self._output-1)*2, self._speed)
         self._outer.setPwm((self._output-1)*2+1, 0)
        else:
          self._outer.setPwm((self._output-1)*2, 0)
          self._outer.setPwm((self._output-1)*2+1, -self._speed)
      def setDistance(self, distance, syncto=None):
        self._outer.setMotorDistance(self._output-1, distance)
        self._distance=distance
        if syncto:
          self._outer.setMotorSyncMaster(self._output-1, syncto._output)
          self._outer.setMotorSyncMaster(syncto._output-1, self._output)
        self._command_id=self._outer.getCurrentMotorCmdId(self._output-1)
        self._outer.incrMotorCmdId(self._output-1)
      def finished(self):
        self._outer.update(wait=True)
        if self._outer.getCurrentMotorCmdId(self._output-1) <= self._command_id:
          return False
        else:
          return True
      def stop(self):
        self.setSpeed(0)
        self.setDistance(0)
        #self._outer.update(wait=True)
    
    M, I = self.getConfig()
    M[output-1] = ftTXT.C_MOTOR
    self.setConfig(M, I)
    self.updateConfig()
    return mot(self, output)

  def output(self, num, level=0):
    """
      Diese Funktion erzeugt ein allgemeines Output-Objekt, das zur Ansteuerung von Elementen verwendet
      wird, die an den Ausgaengen O1-O8 angeschlossen sind.
      
      Anwendungsbeispiel:
      
      Am Ausgang O7 ist eine Lampe oder eine LED angeschlossen:
      
      >>> Lampe = ftrob.output(7)
      
      Das so erzeugte allg. Output-Objekt hat folgende Methoden:
      
      **setLevel** (level)
      
      :param level: Ausgangsleistung, die am Output anliegen soll (genauer fuer die Experten: die Gesamtlaenge des Arbeitsintervalls eines PWM-Taktes in Einheiten von 1/512, d.h. mit level=512 liegt das PWM-Signal waehrend des gesamten Taktes auf high).
      :type level: integer, 1 - 512
      
      Mit dieser Methode kann die Ausgangsleistung eingestellt werden, um z.B. die Helligkeit
      einer Lampe zu regeln.
      
      Anwendungsbeispiel:
      
      Die tatsaechlichen Werte werden erst nach einem update(wait=True) an den TXT uebertragen.
      
      >>> Lampe.setLevel(512)
      >>> ftrob.update(wait=TRUE)
    """
    class out(object):
      def __init__(self, outer, num, level):
        self._outer=outer
        self._num=num
        self._level=level
      def setLevel(self, level):
        self._level=level
        self._outer.setPwm(num-1, self._level)
        #self.outer.update()
    
    M, I = self.getConfig()
    M[(num-1)/2] = ftTXT.C_OUTPUT
    self.setConfig(M, I)
    self.updateConfig()
    return out(self, num, level)    

  def input(self, num):
    """
      Diese Funktion erzeugt ein allgemeines Input-Objekt fuer Sensoren oder Taster,
      die an einem der Eingaenge I1-I8 des TXT angeschlossen sind.
      
      Anwendungsbeispiel:
      
      >>> Taster = ftrob.input(5)
      
      Das so erzeugte allg. Input-Objekt hat folgende Methoden:
      
      **state** ()
      
      Mit dieser Methode wird der Status des Eingangs abgefragt.
      
      :return: Zustand des Schalters (0: Kontakt geschlossen, 1: Kontakt geoeffnet)
      :rtype: integer
      
      Anwendungsbeispiel:
      
      Hinweis: Die update(wait=True) Methode zum Holen der aktuellen Werte vom TXT wird nicht automatisch
               innerhalb der state()-Routine ausgefuehrt. Sie muss bei Bedarf vorher  von Hand ausgefuehrt werden.
      
      >>> ftrob.update(wait=True)
      >>> if Taster.state() == 1:
            print "Der Taster an Eingang I5 wurde gedrueckt."
    """
    class inp(object):
      def __init__(self, outer, num):
        self._outer=outer
        self._num=num
      def state(self):
        return self._outer.getCurrentInput(num-1)
    
    M, I = self.getConfig()
    I[num-1]= (ftTXT.C_SWITCH, ftTXT.C_DIGITAL)
    self.setConfig(M, I)
    self.updateConfig()
    return inp(self, num)
  
  def ultrasonic(self, num):
    """
      Diese Funktion erzeugt ein Objekt zur Abfrage eines an einem der Eingaenge I1-I8 angeschlossenen
      TX-Ultraschall-Distanzmessers.
      
      Anwendungsbeispiel:

      >>> ultraschall = ftrob.ultrasonic(6)
      
      Das so erzeugte Ultraschall-Objekt hat folgende Methoden:
      
      **distance** ()
      
      Mit dieser Methode wird der aktuelle Distanz-Wert abgefragt
      
      :return: Die aktuelle Distanz zwischen Ultraschallsensor und vorgelagertem Objekt in cm.
      :rtype: integer
      
      Hinweis: Die update(wait=True) Methode zum Holen der aktuellen Werte vom TXT wird nicht automatisch
      innerhalb der state()-Routine ausgefuehrt. Sie muss bei Bedarf vorher  von Hand ausgefuehrt werden.
      
      Anwendungsbeispiel:
      
      >>> ftrob.update(wait=True)
      >>> print "Der Abstand zur Wand betraegt ", ultraschall.distance(), " cm."
    """
    class inp(object):
      def __init__(self, outer, num):
        self._outer=outer
        self._num=num
      def distance(self):
        return self._outer.getCurrentInput(num-1)
    
    M, I = self.getConfig()
    I[num-1]= (ftTXT.C_ULTRASONIC, ftTXT.C_ANALOG)
    self.setConfig(M, I)
    self.updateConfig()
    return inp(self, num)

  def sound_finished(self):
    """
      Ueberpruefen, ob die Zeit des zuletzt gespielten Sounds bereits abgelaufen ist
      
      :return: True (Zeit ist abgelaufen) oder False (Sound wird noch abgespielt)
      :rtype: boolean

      Anwendungsbeispiel:
      
      >>> while not ftrob.sound_finished():
            pass
    """
    if time.time()-self._sound_timer > self._sound_length:
      return True
    else:
      return False

  def play_sound(self, idx, seconds=1):
    """
      Einen Sound eine bestimmte zeitlang abspielen
      
      Anwendungsbeispiel:
      
      >>> ftrob.play_sound(27, 5) # 5 Sekunden lang Fahrgeraeusche abspielen
    """
    if time.time()-self._sound_timer > self._sound_length:
      self.setSoundIndex(idx)
      self.setSoundRepeat(1)
      self.incrSoundCmdId()
      self.update(wait=True)
      self.sound_timer=time.time()
      self.sound_length=seconds

  def stop_sound(self):
    """
      Die Aktuelle Soundausgabe stoppen. Dabei wird der abzuspielende Sound-Index auf 0 (=Kein Sound)
      und der Wert fuer die Anzahl der Wiederholungen auf 1 gesetzt.
      
      Hinweis: die Befehle incrSoundCmdId() und update(wait=True) werden innerhalb dieser Routine automatisch
               aufgerufen.
      
      :return: Leer
      
      Anwendungsbeispiel:
      
      >>> ftrob.stop_sound()
    """
    self.setSoundIndex(0)
    self.setSoundRepeat(1)
    self.incrSoundCmdId()
    self.update(wait=True)

